from caddo_file_parser.models.caddo_file import CaddoFile


class CaddoFileValidator:
    def validate(self, caddo_file: CaddoFile):
        print("Making validation...")
        self._validate_run_consistency(caddo_file)
        self._validate_index_set_number(caddo_file)
        self._validate_index_set_counts_in_runs(caddo_file)
        self._validate_seed_consistency_in_run(caddo_file)
        self._settings_objects_set(caddo_file)
        self._is_at_least_one_x_column(caddo_file)
        self._is_at_least_one_y_column(caddo_file)
        self._data_contains_index_colum(caddo_file)
        self._each_index_in_index_sets_points_to_valid_data(caddo_file)
        self._does_runs_contains_valid_numbering(caddo_file)
        self._does_index_sets_contains_valid_numbering(caddo_file)

    def _validate_run_consistency(self, caddo_file: CaddoFile):
        print("Checking if number of Runs in settings is this same as in cadd_file")
        runs_number_in_settings = caddo_file.settings.data_splitting_runs
        runs_number_in_caddo_file = len(caddo_file.runs)
        print(f"Declared in settings: {runs_number_in_settings}, found in caddo_file {runs_number_in_caddo_file}")
        is_valid = runs_number_in_caddo_file == runs_number_in_settings
        if not is_valid:
            raise AttributeError(
                f"There is inconsistent number of Runs, settings declared {runs_number_in_settings} but found {runs_number_in_caddo_file}")
        print()

    def _validate_index_set_counts_in_runs(self, caddo_file: CaddoFile):
        print(
            "Checking if all Runs have this same number of index_sets and if each index_contains this same number of indexes")
        last_number_of_index_sets = len(caddo_file.runs[0].index_sets)
        last_number_of_train_elements_in_index_set = len(caddo_file.runs[0].index_sets[0].train_indexes)
        last_number_of_test_elements_in_index_set = len(caddo_file.runs[0].index_sets[0].test_indexes)
        for run in caddo_file.runs:
            print(f"Checking run number {run.number}:")
            print(f"Number of index_sets found: {len(run.index_sets)}")
            if (len(run.index_sets)) != last_number_of_index_sets:
                raise AttributeError(
                    f"There is inconsistent number of elements in run {run} and in first run was {last_number_of_index_sets}")
            for index_set in run.index_sets:
                if len(index_set.train_indexes) != last_number_of_train_elements_in_index_set:
                    raise AttributeError(
                        f"There is inconsistent number of elements in index_set train indexes {index_set.train_indexes} and in first index set {last_number_of_train_elements_in_index_set}")
                if len(index_set.test_indexes) != last_number_of_test_elements_in_index_set:
                    raise AttributeError(
                        f"There is inconsistent number of elements in index_set test indexes {index_set.test_indexes} and in first index set {last_number_of_test_elements_in_index_set}")
            print("All index sets in this run has this same number of test indexes")
            print("All index sets in this run has this same number of train indexes")
            print()

    def _validate_index_set_number(self, caddo_file):
        print("Checking if each run has declared in settings number of index_sets")
        if caddo_file.settings.data_splitting_folding_method is not None:
            index_set_number_in_settings = caddo_file.settings.data_splitting_folding_number
            min_index_set_number_in_caddo_file = min([len(run.index_sets) for run in caddo_file.runs])
            print(f"Declared number of index_sets in each run is: {index_set_number_in_settings}")
            print(f"Minimal number of index_sets found in run is: {min_index_set_number_in_caddo_file}")
            if min_index_set_number_in_caddo_file != index_set_number_in_settings:
                raise AttributeError(
                    f"There is inconsistent number of index_sets, settings declared {index_set_number_in_settings} but found {min_index_set_number_in_caddo_file} in one of runs")
        print()

    def _validate_seed_consistency_in_run(self, caddo_file: CaddoFile):
        print("Checking if seeds in runs are consistent")
        for run in caddo_file.runs:
            seed_in_run = run.seed
            print(f"Found seed in run number {run.number} is {seed_in_run}")
            for index_set in run.index_sets:
                seed_in_index_set = index_set.seed
                print(f"Found seed in index_set is {seed_in_index_set}")
                if seed_in_run != seed_in_index_set:
                    raise AttributeError(
                        f"Seed in index_set {seed_in_index_set} is different than seed in Run {seed_in_run}")
        print()

    def _settings_objects_set(self, caddo_file: CaddoFile):
        print("Checking if settings object is set")
        if caddo_file.settings is None:
            raise AttributeError(f"Settings property not set")
        print()

    def _is_at_least_one_x_column(self, caddo_file: CaddoFile):
        print("Checking if there is at lest one X column")
        data = caddo_file.data
        x_cols_num = len(data.filter(regex="^[xX]__").columns)
        if x_cols_num == 0:
            raise AttributeError(f"There is no column marked as X (starting with x__)")
        print(f"There is {x_cols_num} X columns")
        print()

    def _is_at_least_one_y_column(self, caddo_file: CaddoFile):
        print("Checking if there is at least one Y column")
        data = caddo_file.data
        y_cols_num = len(data.filter(regex="^[yY]__").columns)
        if y_cols_num == 0:
            raise AttributeError(f"There is no column marked as Y (starting with y__)")
        print(f"There is {y_cols_num} Y columns")
        print()

    def _each_index_in_index_sets_points_to_valid_data(self, caddo_file: CaddoFile):
        print("Checking if each index in index_sets points to valid data in data.csv")
        data = caddo_file.data
        for run in caddo_file.runs:
            for index_set in run.index_sets:
                for index in index_set.test_indexes:
                    try:
                        data.loc[index]
                    except KeyError:
                        raise AttributeError(
                            f"In run {run.number} in test index_set {index_set.number} was index {index} that not match data")
                for index in index_set.train_indexes:
                    try:
                        data.loc[index]
                    except KeyError:
                        raise AttributeError(
                            f"In run {run.number} in train index_set {index_set.number} was index {index} that not match data")
        print("All indexes are OK")
        print()

    def _data_contains_index_colum(self, caddo_file: CaddoFile):
        print("Checking if data.csv contains idx column")
        data = caddo_file.data
        if data["idx"] is None:
            raise AttributeError("Data has no idx column, add idx column")
        print("There is idx column")
        print()

    def _does_runs_contains_valid_numbering(self, caddo_file):
        print("Checking if runs have valid numbering")
        runs = caddo_file.runs
        ids_to_match = list(range(len(runs)))
        for run in runs:
            try:
                ids_to_match.remove(run.number)
            except ValueError:
                raise AttributeError(
                    f"Run number {run.number} is out of range. Those should be in range <0,{len(runs)})")
        if len(ids_to_match) > 0:
            raise AttributeError(f"There was Run ids that was missing {ids_to_match}")
        print("All runs have valid number")
        print()

    def _does_index_sets_contains_valid_numbering(self, caddo_file: CaddoFile):
        for run in caddo_file.runs:
            ids_to_match = list(range(len(run.index_sets)))
            for index_set in run.index_sets:
                try:
                    ids_to_match.remove(index_set.number)
                except ValueError:
                    raise AttributeError(
                        f"Run number {run.number} index_set {index_set.number} is out of range. Those should be in range <0,{len(run.index_sets)})")
            if len(ids_to_match) > 0:
                raise AttributeError(f"There was Run ids that was missing {ids_to_match}")



import unittest

import yaml
from yaml import SafeLoader

from caddo_file_parser.settings.generation_settings_loader import GenerationSettingsLoader


class SettingsLoaderTest(unittest.TestCase):
    def test_load(self):
        settings_loader = GenerationSettingsLoader()
        with open("settings.yaml", 'r') as file:
            settings = settings_loader.load_settings_object(yaml.load(file, Loader=SafeLoader))
            print(settings)



import os
import zipfile
import io

from yaml import SafeLoader

from caddo_file_parser.models.caddo_file import CaddoFile
import pandas as pd
import yaml

from caddo_file_parser.models.index_set import IndexSet
from caddo_file_parser.models.run import Run
from caddo_file_parser.settings.generation_settings_loader import GenerationSettingsLoader
from caddo_file_parser.validation.caddo_file_validator import CaddoFileValidator


class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(Dumper, self).increase_indent(flow, False)


class CaddoFileParser:

    def create_file(self, caddo_file: CaddoFile):
        CaddoFileValidator().validate(caddo_file)
        self.save_data(caddo_file)
        self.save_runs(caddo_file)
        self.save_seeds(caddo_file)
        self.pack_to_caddo_file(caddo_file)
        self.remove_unused_file(caddo_file)

    def save_data(self, caddo_file):
        pd.DataFrame(caddo_file.data).to_csv(
            "data.csv",
            sep=caddo_file.settings.data_output_file_separator,
            index=False
        )

    def save_runs(self, caddo_file):
        for run in caddo_file.runs:
            self.save_index_sets(run)

    def save_seeds(self, caddo_file):
        with open("seeds.yaml", 'w') as file:
            file_content = {
                "seeds": caddo_file.seeds,
            }
            print(file_content)
            yaml.dump(file_content, file, Dumper=Dumper, default_flow_style=False)

    def save_index_sets(self, run):
        for index_set in run.index_sets:
            index_set_number = index_set.number
            train_indexes = index_set.train_indexes
            test_indexes = index_set.test_indexes
            seed = index_set.seed
            file_content = {
                "number": index_set_number,
                "train_indexes": train_indexes,
                "test_indexes": test_indexes,
                "seed": seed
            }
            print(file_content)
            with open(f"index_set_{index_set_number}_run_{run.number}.yaml", 'w') as file:
                yaml.dump(file_content, file, Dumper=Dumper, default_flow_style=False)

    def pack_to_caddo_file(self, caddo_file):
        filenames = []
        for run in caddo_file.runs:
            filenames += [f"index_set_{index_set.number}_run_{run.number}.yaml" for index_set in caddo_file.runs]
        filenames += ["data.csv"] + [caddo_file.settings.data_settings_file_path] + ["seeds.yaml"]
        with zipfile.ZipFile(f"{caddo_file.settings.data_output_file_name}.caddo", "w") as archive:
            for filename in filenames:
                archive.write(filename)

    def remove_unused_file(self, caddo_file):
        for run in caddo_file.runs:
            filenames = [f"index_set_{index_set.number}_run_{run.number}.yaml" for index_set in caddo_file.runs]
            for file in filenames:
                os.remove(file)
        os.remove("data.csv")
        if caddo_file.settings.data_splitting_folding_seeds_file_path != "seeds.yaml" \
                and caddo_file.settings.data_settings_file_path != "./seeds.yaml":
            os.remove("seeds.yaml")

    def read_data(self, file_name) -> CaddoFile:
        with zipfile.ZipFile(file_name + ".caddo", "r") as zf:
            generation_settings = self.read_settings(zf)
            data = self.read_csv_data(zf, generation_settings)
            runs = self.read_runs(zf, generation_settings)
            seeds = self.read_seeds_from_zip(zf, generation_settings)
        caddo_file: CaddoFile = CaddoFile(runs, data, generation_settings, seeds)
        CaddoFileValidator().validate(caddo_file)
        return caddo_file

    def read_settings(self, zf):
        settings_file = zf.read("settings.yaml").decode(encoding="utf-8")
        settings_yaml = yaml.load(settings_file, Loader=SafeLoader)
        return GenerationSettingsLoader().load_settings_object(settings_yaml)

    def read_csv_data(self, zf, generation_settings):
        separator = generation_settings.data_output_file_separator
        data_csv = zf.read("data.csv").decode(encoding="utf-8")
        return pd.read_csv(io.StringIO(data_csv), sep=separator)

    def read_runs(self, zf, generation_settings):
        runs = []
        total_runs_number = generation_settings.data_splitting_runs
        total_index_sets_in_run = self._get_total_index_sets_in_run(generation_settings)
        for run_number in range(total_runs_number):
            index_sets = []
            for fold_number in range(total_index_sets_in_run):
                index_set_file = zf.read(f"index_set_{fold_number}_run_{run_number}.yaml").decode(encoding="utf-8")
                index_set = IndexSet.of(yaml.load(index_set_file, Loader=SafeLoader))
                index_sets.append(index_set)
            runs.append(
                Run(
                    seed=index_sets[0].seed,
                    number=run_number,
                    index_sets=index_sets
                )
            )
        return runs

    def read_seeds_from_zip(self, zf, generation_settings):
        if generation_settings.data_splitting_folding_seeds_from_list:
            return generation_settings.data_splitting_folding_seeds_from_list
        else:
            seeds_file = zf.read(f"{generation_settings.data_splitting_folding_seeds_file_path}").decode(encoding="utf-8")
            return yaml.load(seeds_file, Loader=SafeLoader)["seeds"]

    def read_seeds(self, generation_settings):
        if generation_settings.data_splitting_folding_seeds_file_path != '':
            with open(generation_settings.data_splitting_folding_seeds_file_path) as f:
                return yaml.load(f, Loader=SafeLoader)["seeds"]
        else:
            return generation_settings.data_splitting_folding_seeds_from_list

    def _get_total_index_sets_in_run(self, generation_settings):
        if generation_settings.data_splitting_folding_method is not None:
            return generation_settings.data_splitting_folding_number
        else:
            raise AttributeError("There is no splitting method set")



from caddo_file_parser.settings.generation_settings import GenerationSettings
from caddo_file_parser.validation.settings_validator import SettingsValidator


class GenerationSettingsLoader:
    def load_settings_object(self, settings_file):
        settings_data: GenerationSettings = GenerationSettings()
        if settings_file["data"]["settings_file"]["path"] is not None:
            settings_data.data_settings_file_path = settings_file["data"]["settings_file"]["path"]
        settings_data.data_input_path = settings_file["data"]["input"]["path"]
        settings_data.data_input_separator = settings_file["data"]["input"]["separator"]
        settings_data.data_extraction_function_path = settings_file["data"]['extraction']['function']['path']
        settings_data.data_splitting_folding_number = settings_file["data"]['splitting']['folding']['number']
        settings_data.data_splitting_runs = settings_file["data"]['splitting']['runs']
        settings_data.data_output_file_name = settings_file["data"]['output']['file']['name']
        settings_data.data_splitting_folding_method = settings_file["data"]['splitting']['folding']['method']
        if settings_file["data"]['splitting']['folding']['seeds']['from_list'] is not None:
            settings_data.data_splitting_folding_seeds_from_list = settings_file["data"]['splitting']['folding']['seeds']['from_list']
        if settings_file["data"]['splitting']['folding']['seeds']['from_file'] is not None:
            settings_data.data_splitting_folding_seeds_file_path = settings_file["data"]['splitting']['folding']['seeds']['from_file']
        settings_data.data_output_file_separator = settings_file["data"]['output']['file']['separator']
        SettingsValidator().validate(settings_data)
        return settings_data
